@@ -32,6 +32,20 @@ ghost_actor::ghost_chan! {
         /// Register space event handler
         fn register_space_event_handler(recv: futures::channel::mpsc::Receiver<KitsuneP2pEvent>) -> ();
 
+        /// Incoming Delegate Broadcast
+        /// We are being requested to delegate a broadcast to our neighborhood
+        /// on behalf of an author. `mod_idx` / `mod_cnt` inform us which
+        /// neighbors we are responsible for.
+        /// (See comments in actual method impl for more detail.)
+        fn incoming_delegate_broadcast(
+            space: Arc<KitsuneSpace>,
+            basis: Arc<KitsuneBasis>,
+            to_agent: Arc<KitsuneAgent>,
+            mod_idx: u32,
+            mod_cnt: u32,
+            data: crate::wire::WireData,
+        ) -> ();
+
         /// Incoming Gossip
         fn incoming_gossip(space: Arc<KitsuneSpace>, con: Tx2ConHnd<wire::Wire>, data: Box<[u8]>) -> ();
     }
@@ -143,6 +157,9 @@ impl KitsuneP2pActor {
                 loop {
                     // see if we need a new connection to the proxy
                     if con.is_none() || con.as_ref().unwrap().is_closed() {
+                        if ep_hnd.is_closed() {
+                            break;
+                        }
                         match ep_hnd
                             .get_connection(use_proxy.clone(), tuning_params.implicit_timeout())
                             .await
@@ -159,6 +176,7 @@ impl KitsuneP2pActor {
                     // this is very naive... just running every 5 seconds
                     tokio::time::sleep(std::time::Duration::from_secs(5)).await;
                 }
+                tracing::warn!("proxy con refresh loop shutdown");
             });
 
             ProxyUrl::new(proxy_url.as_base().as_str(), this_digest)
@@ -175,114 +193,144 @@ impl KitsuneP2pActor {
         tokio::task::spawn({
             let evt_sender = evt_sender.clone();
             let tuning_params = config.tuning_params.clone();
-            ep.for_each_concurrent(tuning_params.concurrent_limit_per_thread, move |event| {
-                let evt_sender = evt_sender.clone();
-                let tuning_params = tuning_params.clone();
-                let i_s = i_s.clone();
-                async move {
-                    macro_rules! resp {
-                        ($r:expr, $e:expr) => {
-                            // this can only error as channel closed
-                            // it would be noise to output tracing errors
-                            let _ = $r.respond($e, tuning_params.implicit_timeout()).await;
-                        };
-                    }
+            async move {
+                ep.for_each_concurrent(tuning_params.concurrent_limit_per_thread, move |event| {
+                    let evt_sender = evt_sender.clone();
+                    let tuning_params = tuning_params.clone();
+                    let i_s = i_s.clone();
+                    async move {
+                        macro_rules! resp {
+                            ($r:expr, $e:expr) => {
+                                // this can only error as channel closed
+                                // it would be noise to output tracing errors
+                                let _ = $r.respond($e, tuning_params.implicit_timeout()).await;
+                            };
+                        }
 
-                    let evt_sender = &evt_sender;
-                    use tx2_api::Tx2EpEvent::*;
-                    #[allow(clippy::single_match)]
-                    match event {
-                        IncomingRequest(Tx2EpIncomingRequest { data, respond, .. }) => match data {
-                            wire::Wire::Call(wire::Call {
-                                space,
-                                from_agent,
-                                to_agent,
-                                data,
-                                ..
-                            }) => {
-                                let res = match evt_sender
-                                    .call(space, to_agent, from_agent, data.into())
-                                    .await
-                                {
-                                    Err(err) => {
-                                        let reason = format!("{:?}", err);
-                                        let fail = wire::Wire::failure(reason);
-                                        resp!(respond, fail);
-                                        return;
-                                    }
-                                    Ok(r) => r,
-                                };
-                                let resp = wire::Wire::call_resp(res.into());
-                                resp!(respond, resp);
-                            }
-                            wire::Wire::Notify(wire::Notify {
-                                space,
-                                from_agent,
-                                to_agent,
-                                data,
-                                ..
-                            }) => {
-                                if let Err(err) = evt_sender
-                                    .notify(space, to_agent, from_agent, data.into())
-                                    .await
-                                {
-                                    let reason = format!("{:?}", err);
-                                    let fail = wire::Wire::failure(reason);
-                                    resp!(respond, fail);
-                                    return;
-                                }
-                                let resp = wire::Wire::notify_resp();
-                                resp!(respond, resp);
-                            }
-                            wire::Wire::PeerGet(wire::PeerGet { space, agent }) => {
-                                if let Ok(Some(agent_info_signed)) = evt_sender
-                                    .get_agent_info_signed(GetAgentInfoSignedEvt { space, agent })
-                                    .await
-                                {
-                                    let resp = wire::Wire::peer_get_resp(agent_info_signed);
-                                    resp!(respond, resp);
-                                } else {
-                                    let resp = wire::Wire::failure("no such agent".into());
-                                    resp!(respond, resp);
-                                }
-                            }
-                            wire::Wire::PeerQuery(wire::PeerQuery { space, basis_loc }) => {
-                                // this *does* go over the network...
-                                // so we don't want it to be too many
-                                const LIMIT: u32 = 8;
-                                match evt_sender
-                                    .query_agent_info_signed_near_basis(space, basis_loc, LIMIT)
-                                    .await
-                                {
-                                    Ok(list) if !list.is_empty() => {
-                                        let resp = wire::Wire::peer_query_resp(list);
+                        let evt_sender = &evt_sender;
+                        use tx2_api::Tx2EpEvent::*;
+                        #[allow(clippy::single_match)]
+                        match event {
+                            IncomingRequest(Tx2EpIncomingRequest { data, respond, .. }) => {
+                                match data {
+                                    wire::Wire::Call(wire::Call {
+                                        space,
+                                        from_agent,
+                                        to_agent,
+                                        data,
+                                        ..
+                                    }) => {
+                                        let res = match evt_sender
+                                            .call(space, to_agent, from_agent, data.into())
+                                            .await
+                                        {
+                                            Err(err) => {
+                                                let reason = format!("{:?}", err);
+                                                let fail = wire::Wire::failure(reason);
+                                                resp!(respond, fail);
+                                                return;
+                                            }
+                                            Ok(r) => r,
+                                        };
+                                        let resp = wire::Wire::call_resp(res.into());
                                         resp!(respond, resp);
                                     }
-                                    res => {
-                                        let resp = wire::Wire::failure(format!(
-                                            "error getting agents: {:?}",
-                                            res
-                                        ));
-                                        resp!(respond, resp);
+                                    wire::Wire::PeerGet(wire::PeerGet { space, agent }) => {
+                                        if let Ok(Some(agent_info_signed)) = evt_sender
+                                            .get_agent_info_signed(GetAgentInfoSignedEvt {
+                                                space,
+                                                agent,
+                                            })
+                                            .await
+                                        {
+                                            let resp = wire::Wire::peer_get_resp(agent_info_signed);
+                                            resp!(respond, resp);
+                                        } else {
+                                            let resp = wire::Wire::failure("no such agent".into());
+                                            resp!(respond, resp);
+                                        }
+                                    }
+                                    wire::Wire::PeerQuery(wire::PeerQuery { space, basis_loc }) => {
+                                        // this *does* go over the network...
+                                        // so we don't want it to be too many
+                                        const LIMIT: u32 = 8;
+                                        match evt_sender
+                                            .query_agent_info_signed_near_basis(
+                                                space, basis_loc, LIMIT,
+                                            )
+                                            .await
+                                        {
+                                            Ok(list) if !list.is_empty() => {
+                                                let resp = wire::Wire::peer_query_resp(list);
+                                                resp!(respond, resp);
+                                            }
+                                            res => {
+                                                let resp = wire::Wire::failure(format!(
+                                                    "error getting agents: {:?}",
+                                                    res
+                                                ));
+                                                resp!(respond, resp);
+                                            }
+                                        }
+                                    }
+                                    data => unimplemented!("{:?}", data),
+                                }
+                            }
+                            IncomingNotify(Tx2EpIncomingNotify { con, data, .. }) => match data {
+                                wire::Wire::DelegateBroadcast(wire::DelegateBroadcast {
+                                    space,
+                                    basis,
+                                    to_agent,
+                                    mod_idx,
+                                    mod_cnt,
+                                    data,
+                                }) => {
+                                    // one might be tempted to notify here
+                                    // as in Broadcast below... but we
+                                    // notify all relevent agents inside
+                                    // the space incoming_delegate_broadcast
+                                    // handler.
+                                    if let Err(err) = i_s
+                                        .incoming_delegate_broadcast(
+                                            space, basis, to_agent, mod_idx, mod_cnt, data,
+                                        )
+                                        .await
+                                    {
+                                        tracing::warn!(
+                                            ?err,
+                                            "failed to handle incoming delegate broadcast"
+                                        );
                                     }
                                 }
-                            }
-                            data => unimplemented!("{:?}", data),
-                        },
-                        IncomingNotify(Tx2EpIncomingNotify { con, data, .. }) => match data {
-                            wire::Wire::Gossip(wire::Gossip { space, data }) => {
-                                let data: Vec<u8> = data.into();
-                                let data: Box<[u8]> = data.into_boxed_slice();
-                                if let Err(e) = i_s.incoming_gossip(space, con, data).await {
-                                    tracing::warn!("failed to handle incoming gossip: {:?}", e);
+                                wire::Wire::Broadcast(wire::Broadcast {
+                                    space,
+                                    to_agent,
+                                    data,
+                                    ..
+                                }) => {
+                                    if let Err(err) = evt_sender
+                                        .notify(space, to_agent.clone(), to_agent, data.into())
+                                        .await
+                                    {
+                                        tracing::warn!(?err, "error processing incoming broadcast");
+                                    }
                                 }
-                            }
-                            data => unimplemented!("{:?}", data),
-                        },
-                        _ => (),
+                                wire::Wire::Gossip(wire::Gossip { space, data }) => {
+                                    let data: Vec<u8> = data.into();
+                                    let data: Box<[u8]> = data.into_boxed_slice();
+                                    if let Err(e) = i_s.incoming_gossip(space, con, data).await {
+                                        tracing::warn!("failed to handle incoming gossip: {:?}", e);
+                                    }
+                                }
+                                data => unimplemented!("{:?}", data),
+                            },
+                            _ => (),
+                        }
                     }
-                }
-            })
+                })
+                .await;
+                tracing::warn!("KitsuneP2p tx2:ep poll shutdown");
+            }
         });
 
         Ok(Self {
@@ -297,7 +345,24 @@ impl KitsuneP2pActor {
     }
 }
 
-impl ghost_actor::GhostControlHandler for KitsuneP2pActor {}
+use ghost_actor::dependencies::must_future::MustBoxFuture;
+impl ghost_actor::GhostControlHandler for KitsuneP2pActor {
+    fn handle_ghost_actor_shutdown(mut self) -> MustBoxFuture<'static, ()> {
+        use futures::sink::SinkExt;
+        use ghost_actor::GhostControlSender;
+        async move {
+            // this is a curtesy, ok if fails
+            let _ = self.evt_sender.close().await;
+            self.ep_hnd.close(500, "").await;
+            for (_, space) in self.spaces.into_iter() {
+                let (space, _) = space.get().await;
+                let _ = space.ghost_actor_shutdown_immediate().await;
+            }
+        }
+        .boxed()
+        .into()
+    }
+}
 
 impl ghost_actor::GhostHandler<Internal> for KitsuneP2pActor {}
 
@@ -315,6 +380,35 @@ impl InternalHandler for KitsuneP2pActor {
         .into())
     }
 
+    fn handle_incoming_delegate_broadcast(
+        &mut self,
+        space: Arc<KitsuneSpace>,
+        basis: Arc<KitsuneBasis>,
+        to_agent: Arc<KitsuneAgent>,
+        mod_idx: u32,
+        mod_cnt: u32,
+        data: crate::wire::WireData,
+    ) -> InternalHandlerResult<()> {
+        let space_sender = match self.spaces.get_mut(&space) {
+            None => {
+                tracing::warn!(
+                    "received delegate_broadcast for unhandled space: {:?}",
+                    space
+                );
+                return Ok(async move { Ok(()) }.boxed().into());
+            }
+            Some(space) => space.get(),
+        };
+        Ok(async move {
+            let (_, space_inner) = space_sender.await;
+            space_inner
+                .incoming_delegate_broadcast(space, basis, to_agent, mod_idx, mod_cnt, data)
+                .await
+        }
+        .boxed()
+        .into())
+    }
+
     fn handle_incoming_gossip(
         &mut self,
         space: Arc<KitsuneSpace>,
@@ -535,14 +629,20 @@ impl KitsuneP2pHandler for KitsuneP2pActor {
         .into())
     }
 
-    fn handle_notify_multi(&mut self, input: actor::NotifyMulti) -> KitsuneP2pHandlerResult<u8> {
-        let space_sender = match self.spaces.get_mut(&input.space) {
-            None => return Err(KitsuneP2pError::RoutingSpaceError(input.space)),
+    fn handle_broadcast(
+        &mut self,
+        space: Arc<KitsuneSpace>,
+        basis: Arc<KitsuneBasis>,
+        timeout: KitsuneTimeout,
+        payload: Vec<u8>,
+    ) -> KitsuneP2pHandlerResult<()> {
+        let space_sender = match self.spaces.get_mut(&space) {
+            None => return Err(KitsuneP2pError::RoutingSpaceError(space)),
             Some(space) => space.get(),
         };
         Ok(async move {
             let (space_sender, _) = space_sender.await;
-            space_sender.notify_multi(input).await
+            space_sender.broadcast(space, basis, timeout, payload).await
         }
         .boxed()
         .into())
