@@ -27,6 +27,20 @@ ghost_actor::ghost_chan! {
         /// see if an agent is locally joined
         fn is_agent_local(agent: Arc<KitsuneAgent>) -> bool;
 
+        /// Incoming Delegate Broadcast
+        /// We are being requested to delegate a broadcast to our neighborhood
+        /// on behalf of an author. `mod_idx` / `mod_cnt` inform us which
+        /// neighbors we are responsible for.
+        /// (See comments in actual method impl for more detail.)
+        fn incoming_delegate_broadcast(
+            space: Arc<KitsuneSpace>,
+            basis: Arc<KitsuneBasis>,
+            to_agent: Arc<KitsuneAgent>,
+            mod_idx: u32,
+            mod_cnt: u32,
+            data: crate::wire::WireData,
+        ) -> ();
+
         /// Incoming Gossip
         fn incoming_gossip(space: Arc<KitsuneSpace>, con: Tx2ConHnd<wire::Wire>, data: Box<[u8]>) -> ();
     }
@@ -126,6 +140,7 @@ impl SpaceInternalHandler for Space {
         .boxed()
         .into())
     }
+
     fn handle_update_single_agent_info(
         &mut self,
         agent: Arc<KitsuneAgent>,
@@ -164,6 +179,87 @@ impl SpaceInternalHandler for Space {
         Ok(async move { Ok(res) }.boxed().into())
     }
 
+    fn handle_incoming_delegate_broadcast(
+        &mut self,
+        space: Arc<KitsuneSpace>,
+        basis: Arc<KitsuneBasis>,
+        _to_agent: Arc<KitsuneAgent>,
+        mod_idx: u32,
+        mod_cnt: u32,
+        data: crate::wire::WireData,
+    ) -> InternalHandlerResult<()> {
+        // first, forward this incoming broadcast to all connected
+        // local agents.
+        let mut local_events = Vec::new();
+        for agent in self.local_joined_agents.iter().cloned() {
+            let fut = self.evt_sender.notify(
+                space.clone(),
+                agent.clone(),
+                agent.clone(),
+                data.clone().into(),
+            );
+            local_events.push(async move {
+                if let Err(err) = fut.await {
+                    tracing::warn!(?err, "failed local broadcast");
+                }
+            });
+        }
+
+        // next, gather a list of agents covering this data to be
+        // published to.
+        let ro_inner = self.ro_inner.clone();
+        let timeout = ro_inner.config.tuning_params.implicit_timeout();
+        let fut =
+            discover::get_cached_remotes_near_basis(ro_inner.clone(), basis.get_loc(), timeout);
+
+        Ok(async move {
+            futures::future::join_all(local_events).await;
+
+            let info_list = fut.await?;
+
+            // for all agents in the gathered list, check the modulo params
+            // i.e. if `agent.get_loc() % mod_cnt == mod_idx` we know we are
+            // responsible for delegating the broadcast to that agent.
+            let mut all = Vec::new();
+            for info in info_list
+                .into_iter()
+                .filter(|info| info.agent.get_loc() % mod_cnt == mod_idx)
+            {
+                let ro_inner = ro_inner.clone();
+                let space = space.clone();
+                let basis = basis.clone();
+                let data = data.clone();
+                all.push(async move {
+                    use discover::PeerDiscoverResult;
+
+                    // attempt to establish a connection
+                    let con_hnd = match discover::peer_connect(ro_inner, &info, timeout).await {
+                        PeerDiscoverResult::OkShortcut => return,
+                        PeerDiscoverResult::OkRemote { con_hnd, .. } => con_hnd,
+                        PeerDiscoverResult::Err(err) => {
+                            tracing::warn!(?err, "broadcast error");
+                            return;
+                        }
+                    };
+
+                    // generate our broadcast payload
+                    let payload = wire::Wire::broadcast(space, basis, info.agent.clone(), data);
+
+                    // forward the data
+                    if let Err(err) = con_hnd.notify(&payload, timeout).await {
+                        tracing::warn!(?err, "broadcast error");
+                    }
+                })
+            }
+
+            futures::future::join_all(all).await;
+
+            Ok(())
+        }
+        .boxed()
+        .into())
+    }
+
     fn handle_incoming_gossip(
         &mut self,
         _space: Arc<KitsuneSpace>,
@@ -185,6 +281,7 @@ struct UpdateAgentInfoInput<'borrow> {
     mdns_handles: &'borrow mut HashMap<Vec<u8>, Arc<AtomicBool>>,
     bootstrap_service: &'borrow Option<Url2>,
 }
+
 async fn update_single_agent_info(input: UpdateAgentInfoInput<'_>) -> KitsuneP2pResult<()> {
     let UpdateAgentInfoInput {
         expires_after,
@@ -267,7 +364,19 @@ async fn update_single_agent_info(input: UpdateAgentInfoInput<'_>) -> KitsuneP2p
     Ok(())
 }
 
-impl ghost_actor::GhostControlHandler for Space {}
+use ghost_actor::dependencies::must_future::MustBoxFuture;
+impl ghost_actor::GhostControlHandler for Space {
+    fn handle_ghost_actor_shutdown(mut self) -> MustBoxFuture<'static, ()> {
+        async move {
+            use futures::sink::SinkExt;
+            // this is a curtesy, ok if fails
+            let _ = self.evt_sender.close().await;
+            self.gossip_mod.close();
+        }
+        .boxed()
+        .into()
+    }
+}
 
 impl ghost_actor::GhostHandler<KitsuneP2p> for Space {}
 
@@ -349,8 +458,8 @@ impl KitsuneP2pHandler for Space {
         agent: Arc<KitsuneAgent>,
     ) -> KitsuneP2pHandlerResult<()> {
         self.local_joined_agents.remove(&agent);
-        self.gossip_mod.local_agent_leave(agent);
-        Ok(async move { Ok(()) }.boxed().into())
+        self.gossip_mod.local_agent_leave(agent.clone());
+        self.publish_leave_agent_info(agent)
     }
 
     fn handle_rpc_single(
@@ -443,43 +552,124 @@ impl KitsuneP2pHandler for Space {
         self.handle_rpc_multi_inner(input)
     }
 
-    fn handle_notify_multi(
+    fn handle_broadcast(
         &mut self,
-        mut input: actor::NotifyMulti,
-    ) -> KitsuneP2pHandlerResult<u8> {
-        // if the user doesn't care about remote_agent_count, apply default
-        match input.remote_agent_count {
-            None | Some(0) => {
-                input.remote_agent_count =
-                    Some(self.config.tuning_params.default_notify_remote_agent_count as u8);
-            }
-            _ => {}
+        space: Arc<KitsuneSpace>,
+        basis: Arc<KitsuneBasis>,
+        timeout: KitsuneTimeout,
+        payload: Vec<u8>,
+    ) -> KitsuneP2pHandlerResult<()> {
+        // first, forward this data to all connected local agents.
+        let mut local_events = Vec::new();
+        for agent in self.local_joined_agents.iter().cloned() {
+            let fut = self.evt_sender.notify(
+                space.clone(),
+                agent.clone(),
+                agent.clone(),
+                payload.clone(),
+            );
+            local_events.push(async move {
+                if let Err(err) = fut.await {
+                    tracing::warn!(?err, "failed local broadcast");
+                }
+            });
         }
 
-        // if the user doesn't care about timeout_ms, apply default
-        // also - if set to 0, we want to return immediately, but
-        // spawn a task with that default timeout.
-        let do_spawn = match input.timeout_ms {
-            None | Some(0) => {
-                input.timeout_ms = Some(self.config.tuning_params.default_notify_timeout_ms as u64);
-                true
-            }
-            _ => false,
-        };
+        // then, find a list of agents in a potentially remote neighborhood
+        // that should be responsible for holding the data.
+        let ro_inner = self.ro_inner.clone();
+        let discover_fut =
+            discover::search_remotes_covering_basis(ro_inner.clone(), basis.get_loc(), timeout);
+        Ok(async move {
+            futures::future::join_all(local_events).await;
 
-        // gather the inner future
-        let inner_fut = match self.handle_notify_multi_inner(input) {
-            Err(e) => return Err(e),
-            Ok(f) => f,
-        };
+            // TODO - FIXME
+            // Holochain currently does all its testing without any remote nodes
+            // if we do this inline, it takes us to the 30 second timeout
+            // on every one of those... so spawning for now, which means
+            // we won't get notified if we are unable to publish to anyone.
+            // Also, if conductor spams us with publishes, we could fill
+            // the memory with publish tasks.
+            tokio::task::spawn(async move {
+                let cover_nodes = discover_fut.await?;
+                if cover_nodes.is_empty() {
+                    return Err("failed to discover neighboring peers".into());
+                }
 
-        // either spawn or return the future depending on timeout_ms logic
-        if do_spawn {
-            tokio::task::spawn(inner_fut);
-            Ok(async move { Ok(0) }.boxed().into())
-        } else {
-            Ok(inner_fut)
+                let mut all = Vec::new();
+
+                // is there a better way to do this??
+                //
+                // since we're gathering the connections in one place,
+                // if any of them take the full timeout, we won't have any
+                // time to actually forward the message to them.
+                //
+                // and if a node is that slow anyways, maybe we don't want
+                // to trust them to forward the message in any case...
+                let half_timeout =
+                    KitsuneTimeout::from_millis(timeout.time_remaining().as_millis() as u64 / 2);
+
+                // attempt to open connections to the discovered remote nodes
+                for info in cover_nodes {
+                    let ro_inner = ro_inner.clone();
+                    all.push(async move {
+                        use discover::PeerDiscoverResult;
+                        let con_hnd =
+                            match discover::peer_connect(ro_inner, &info, half_timeout).await {
+                                PeerDiscoverResult::OkShortcut => return None,
+                                PeerDiscoverResult::OkRemote { con_hnd, .. } => con_hnd,
+                                PeerDiscoverResult::Err(_) => return None,
+                            };
+                        Some((info.agent.clone(), con_hnd))
+                    });
+                }
+
+                let con_list = futures::future::join_all(all)
+                    .await
+                    .into_iter()
+                    .flatten()
+                    .collect::<Vec<_>>();
+
+                if con_list.is_empty() {
+                    return Err("failed to connect to neighboring peers".into());
+                }
+
+                let mut all = Vec::new();
+
+                // determine the total number of nodes we'll be publishing to
+                // we'll make each remote responsible for a subset of delegate
+                // broadcasting by having them apply the formula:
+                // `agent.get_loc() % mod_cnt == mod_idx` -- if true,
+                // they'll be responsible for forwarding the data to that node.
+                let mod_cnt = con_list.len();
+                for (mod_idx, (agent, con_hnd)) in con_list.into_iter().enumerate() {
+                    // build our delegate message
+                    let payload = wire::Wire::delegate_broadcast(
+                        space.clone(),
+                        basis.clone(),
+                        agent,
+                        mod_idx as u32,
+                        mod_cnt as u32,
+                        payload.clone().into(),
+                    );
+
+                    // notify the remote node
+                    all.push(async move {
+                        if let Err(err) = con_hnd.notify(&payload, timeout).await {
+                            tracing::warn!(?err, "delegate broadcast error");
+                        }
+                    });
+                }
+
+                futures::future::join_all(all).await;
+
+                KitsuneP2pResult::Ok(())
+            });
+
+            Ok(())
         }
+        .boxed()
+        .into())
     }
 }
 
@@ -502,7 +692,6 @@ pub(crate) struct Space {
     pub(crate) this_addr: url2::Url2,
     pub(crate) i_s: ghost_actor::GhostSender<SpaceInternal>,
     pub(crate) evt_sender: futures::channel::mpsc::Sender<KitsuneP2pEvent>,
-    pub(crate) ep_hnd: Tx2EpHnd<wire::Wire>,
     pub(crate) local_joined_agents: HashSet<Arc<KitsuneAgent>>,
     pub(crate) config: Arc<KitsuneP2pConfig>,
     mdns_handles: HashMap<Vec<u8>, Arc<AtomicBool>>,
@@ -558,6 +747,11 @@ impl Space {
                 let mut delay_len = START_DELAY;
 
                 loop {
+                    use ghost_actor::GhostControlSender;
+                    if !i_s_c.ghost_actor_is_active() {
+                        break;
+                    }
+
                     tokio::time::sleep(delay_len).await;
                     if delay_len <= MAX_DELAY {
                         delay_len *= 2;
@@ -576,6 +770,9 @@ impl Space {
                             tracing::error!(msg = "Failed to get peers from bootstrap", ?e);
                         }
                         Ok(list) => {
+                            if !i_s_c.ghost_actor_is_active() {
+                                break;
+                            }
                             for item in list {
                                 // TODO - someday some validation here
                                 let agent = item.agent.clone();
@@ -604,6 +801,7 @@ impl Space {
                         }
                     }
                 }
+                tracing::warn!("bootstrap fetch loop ending");
             });
         }
 
@@ -612,7 +810,7 @@ impl Space {
             this_addr: this_addr.clone(),
             i_s: i_s.clone(),
             evt_sender: evt_sender.clone(),
-            ep_hnd: ep_hnd.clone(),
+            ep_hnd,
             config: config.clone(),
         });
 
@@ -622,7 +820,6 @@ impl Space {
             this_addr,
             i_s,
             evt_sender,
-            ep_hnd,
             local_joined_agents: HashSet::new(),
             config,
             mdns_handles: HashMap::new(),
@@ -631,6 +828,72 @@ impl Space {
         }
     }
 
+    fn publish_leave_agent_info(
+        &mut self,
+        agent: Arc<KitsuneAgent>,
+    ) -> KitsuneP2pHandlerResult<()> {
+        let space = self.space.clone();
+        let network_type = self.config.network_type.clone();
+        let evt_sender = self.evt_sender.clone();
+        let bootstrap_service = self.config.bootstrap_service.clone();
+        let expires_after = self.config.tuning_params.agent_info_expires_after_ms as u64;
+        Ok(async move {
+            let signed_at_ms = crate::spawn::actor::bootstrap::now_once(None).await?;
+            let expires_at_ms = signed_at_ms + expires_after;
+            use kitsune_p2p_types::agent_info::AgentInfoSigned;
+            let agent_info_signed = AgentInfoSigned::sign(
+                space.clone(),
+                agent.clone(),
+                0,          // no storage arc
+                Vec::new(), // no urls
+                signed_at_ms,
+                expires_at_ms,
+                |d| {
+                    let data = Arc::new(d.to_vec());
+                    async {
+                        let sign_req = SignNetworkDataEvt {
+                            space: space.clone(),
+                            agent: agent.clone(),
+                            data,
+                        };
+                        evt_sender
+                            .sign_network_data(sign_req)
+                            .await
+                            .map(Arc::new)
+                            .map_err(KitsuneError::other)
+                    }
+                },
+            )
+            .await?;
+
+            tracing::debug!(?agent_info_signed);
+
+            evt_sender
+                .put_agent_info_signed(PutAgentInfoSignedEvt {
+                    space: space.clone(),
+                    agent: agent.clone(),
+                    agent_info_signed: agent_info_signed.clone(),
+                })
+                .await?;
+
+            // Push to the network as well
+            match network_type {
+                NetworkType::QuicMdns => tracing::warn!("NOT publishing leaves to mdns"),
+                NetworkType::QuicBootstrap => {
+                    crate::spawn::actor::bootstrap::put(
+                        bootstrap_service.clone(),
+                        agent_info_signed,
+                    )
+                    .await?;
+                }
+            }
+
+            Ok(())
+        }
+        .boxed()
+        .into())
+    }
+
     /// actual logic for handle_rpc_multi ...
     /// the top-level handler may or may not spawn a task for this
     #[tracing::instrument(skip(self, input))]
@@ -728,66 +991,4 @@ impl Space {
         .boxed()
         .into())
     }
-
-    /// actual logic for handle_notify_multi ...
-    /// the top-level handler may or may not spawn a task for this
-    fn handle_notify_multi_inner(
-        &mut self,
-        input: actor::NotifyMulti,
-    ) -> KitsuneP2pHandlerResult<u8> {
-        let actor::NotifyMulti {
-            space,
-            from_agent,
-            basis,
-            remote_agent_count,
-            timeout_ms,
-            payload,
-        } = input;
-
-        let remote_agent_count = remote_agent_count.expect("set by handle_notify_multi");
-        let timeout_ms = timeout_ms.expect("set by handle_notify_multi");
-        let stage_1_timeout_ms = timeout_ms / 2;
-
-        // as an optimization - broadcast to all local joins
-        // but don't count that toward our publish total
-        let local_all = self
-            .local_joined_agents
-            .iter()
-            .map(|agent| {
-                self.evt_sender.notify(
-                    space.clone(),
-                    agent.clone(),
-                    from_agent.clone(),
-                    payload.clone(),
-                )
-            })
-            .collect::<Vec<_>>();
-
-        let remote_fut = discover::message_neighborhood(
-            self,
-            from_agent.clone(),
-            remote_agent_count,
-            stage_1_timeout_ms,
-            timeout_ms,
-            basis,
-            wire::Wire::notify(
-                space.clone(),
-                from_agent.clone(),
-                from_agent,
-                payload.into(),
-            ),
-            |_, w| match w {
-                wire::Wire::NotifyResp(_) => Ok(()),
-                _ => Err(()),
-            },
-        );
-
-        Ok(async move {
-            futures::future::try_join_all(local_all).await?;
-
-            Ok(remote_fut.await.len() as u8)
-        }
-        .boxed()
-        .into())
-    }
 }
